/**
 * Smith-Waterman algoritmus implementacioja.
 * @author Dombai Balazs
 */
public class SmithWaterman {
    
    private String str1, str2;
    private int[][] d2;
    private int max1, max2;
    
    /**
     * Az elso szekvencia illesztve.
     */
    public String aligned_seq1 = "";
    /**
     * A masodik szekvencia illesztve.
     */
    public String aligned_seq2 = "";
    
    /**
     * Konstruktor.
     * @param str1 elso szekvencia
     * @param str2 masodik szekvencia
     * @param penalty a bunteteseket szamolo osztaly
     */
    public SmithWaterman(String str1, String str2, IPenalty penalty) {
	this.str1 = str1;
	this.str2 = str2;
	
	penalty.init(str1, str2);
	penalty.run();
	penalty.findMax();
	
	d2 = penalty.getDirectionMatrix();
	max1 = penalty.get_max1();
	max2 = penalty.get_max2();
	getOptimalWay();
    }
    
    /**
     * Optimalis ut keresese a jobb also cellatol a bal felso 0-ig.
     */
    public void getOptimalWay() {
	int current1 = max1, current2 = max2;
	while (d2[current1][current2] != IPenalty.STOP) {
	    if (d2[current1][current2] == IPenalty.UPLEFT) {
		current1--;
		current2--;
		aligned_seq1 = str1.charAt(current1) + aligned_seq1;
		aligned_seq2 = str2.charAt(current2) + aligned_seq2;
	    } else if (d2[current1][current2] == IPenalty.UP) {
		current1--;
		aligned_seq1 = str1.charAt(current1) + aligned_seq1;
		aligned_seq2 = "-" + aligned_seq2;
	    } else if (d2[current1][current2] == IPenalty.LEFT) {
		current2--;
		aligned_seq1 = "-" + aligned_seq1;
		aligned_seq2 = str2.charAt(current2) + aligned_seq2;
	    }
	}
    }
    
    /**
     * Az optimalis ut altal meghatarozott (t/r/mu/lambda) negyes lekerdezese.
     * @return (t/r/mu/lambda)
     */
    public Params getParams() {
	return new Params(get_t(), get_r(), get_lambda(), str1.length() + str2.length());
    }
    
    /**
     * Meghatarozza a kezdeti becslest t-re.
     * "Az optimalis illesztesben szamoljuk ossze, hogy hany
     * helyen van egyezes, ezt osszuk le a ket szekvencia hosszanak az
     * osszegevel, majd ennek vegyuk a termeszetes alapu logaritmusat, es ennek
     * a -1-szereset. Ez lesz a kezdeti becslesunk t-re."
     * @return t becsult erteke
     */
    private double get_t() {
	int nummatches = 0;
	for (int i = 0; i < aligned_seq1.length(); i++) {
	    if (aligned_seq1.charAt(i) == aligned_seq2.charAt(i)) {
		nummatches++;
	    }
	}
	double result = -Math.log( (double)nummatches * 2.0 / (str1.length()+str2.length()) );
	if (result == 0.0) result = (double)1 / (double)Integer.MAX_VALUE;
	return result;
    }
    
    /**
     * Meghatarozza a kezdeti becslest r-re.
     * "Nezzuk meg, hogy mi a beszurasok es torlesek atlagos hossza (Istvan: +1...), ennek a
     * reciprokat vonjuk ki 1-bol, ez lesz a kezdeti becslesunk r-re."
     * @return r becsult erteke
     */
    private double get_r() {
	int length = 1;
	for (int i = 0; i < aligned_seq1.length(); i++) {
	    if (aligned_seq1.charAt(i) == '-' || aligned_seq2.charAt(i) == '-') {
		length++;
	    }
	}
	int numops = 0;
	for (int i = 0; i < aligned_seq1.length(); i++) {
	    if ( aligned_seq1.charAt(i) == '-' && (i == 0 || aligned_seq1.charAt(i - 1) != '-') ) {
		numops++;
	    }
	    if ( aligned_seq2.charAt(i) == '-' && (i == 0 || aligned_seq2.charAt(i - 1) != '-') ) {
		numops++;
	    }
	}
	return 1-(double)numops/length;
    }
    
    /**
     * Meghatarozza a kezdeti becslest lambda-ra.
     * "Nezzuk meg, hogy hany beszuras tortent (egy hosszu beszurast egy
     * beszurasnak szamolva), ezt osszuk le a lenti szekvencia
     * hosszaval, valamint a mar becsult t-vel, ez lesz a kezdeti parameter
     * lambdara,"
     * @return lambda kezdeti erteke
     */
    private double get_lambda() {
	int numinsert = 0;
	for (int i = 0; i < aligned_seq1.length(); i++) {
	    if ( aligned_seq1.charAt(i) == '-' && (i == 0 || aligned_seq1.charAt(i - 1) != '-') ) {
		numinsert++;
	    }
	    if ( aligned_seq2.charAt(i) == '-' && (i == 0 || aligned_seq2.charAt(i - 1) != '-') ) {
		numinsert++;
	    }
	    
	}
	return (double)numinsert/(str1.length() + str2.length())/get_t();
    }
    
}
